@@ -1,3 +1,4 @@
+{-# LANGUAGE TemplateHaskell #-}
 ------------------------------------------------------------------------------
 -- |
 -- Module:      Database.PostgreSQL.Simple.SqlQQ
@@ -9,15 +10,16 @@
 ------------------------------------------------------------------------------
 
 module Database.PostgreSQL.Simple.SqlQQ (sql) where
-
+import Database.PostgreSQL.Simple.Types (Query)
 import Language.Haskell.TH
 import Language.Haskell.TH.Quote
 import Data.Char
+import Data.String
 
 -- | 'sql' is a quasiquoter that eases the syntactic burden
 -- of writing big sql statements in Haskell source code.  For example:
 --
--- > {-# LANGUAGE OverloadedStrings, QuasiQuotes #-}
+-- > {-# LANGUAGE QuasiQuotes #-}
 -- >
 -- > query conn [sql| SELECT column_a, column_b
 -- >                    FROM table1 NATURAL JOIN table2
@@ -27,10 +29,10 @@ import Data.Char
 -- >                   LIMIT 100                        |]
 -- >            (beginTime,endTime,string)
 --
--- This quasiquoter attempts to mimimize whitespace;  otherwise the
--- above query would consist of approximately half whitespace when sent
--- to the database backend.  It also recognizes and strips out standard
--- sql comments "--".
+-- This quasiquoter returns a literal string expression of type 'Query',
+-- and attempts to mimimize whitespace;  otherwise the above query would
+-- consist of approximately half whitespace when sent to the database
+-- backend.  It also recognizes and strips out standard sql comments "--".
 --
 -- The implementation of the whitespace reducer is currently incomplete.
 -- Thus it can mess up your syntax in cases where whitespace should be
@@ -60,7 +62,7 @@ sql = QuasiQuoter
     }
 
 sqlExp :: String -> Q Exp
-sqlExp = stringE . minimizeSpace
+sqlExp = appE [| fromString :: String -> Query |] . stringE . minimizeSpace
 
 minimizeSpace :: String -> String
 minimizeSpace = drop 1 . reduceSpace
